//====================================================================================//
//                                                                                    //
//                            Class Probability Vector                                //
//                                                                                    //
//====================================================================================//
//  This File:   vct.C                      Language: C++  (xlC and CC)               //
//  Software:    PEPS 3.1                    Advisor: Brigitte.Plateau@imag.fr        //
//  Error Codes: 2350                                                                 //
//====================================================================================//
//  Creation:    19/dec/95                        by: Paulo.Fernandes@imag.fr         //
//  Last Change: 14/feb/00                        by: Paulo.Fernandes@imag.fr         //
//====================================================================================//
#include "peps.h"

     //--------------------------------------------------------------------------//
     //                 static variable: The Probability Vector                  //
     //--------------------------------------------------------------------------//
VCT VCT::vct = VCT();

     //--------------------------------------------------------------------------//
     //               It initializes the states of 'this' - HBF                  //
     //--------------------------------------------------------------------------//
     //    This function is used to initialize the elements of 'this' according  //
     // to one of the following options:                                         //
     //                                                                          //
     //   - 'eq_vec',      all reachable states are initialized with an uniform  //
     //                    distribution (equiprobable);                          //
     //   - 'ap_vec',      all reachable states are initialized with a           //
     //                    distribution proportional to the inverse of the       //
     //                    diagonal elements of the HBF matrix ('HBF::hbf');     //
     //                                                                          //
     //   The size of the vector is to the HBF matrix order. The fields 'sizes'  //
     // and 'permutation' are erased.                                            //
     //                                                                          //
     // Note: this functions is used to prepare a vector to be multiplied by the //
     // Hardwell-Boeing Format sparse matrix (static variable 'HBF::hbf').       //
     //--------------------------------------------------------------------------//
     // Implementation note: The generation of HBF matrices in PEPS eliminates   //
     // all non-reachable states. Therefore, the order of the HBF matrix is the  //
     // number of reachable states.                                              //
     //--------------------------------------------------------------------------//
void VCT::init_hbf(const v_typ o)
{
 int i;             // a counter for the states
 rp  factor;        // a factor to normalize the vector elements
 rp  acc = zero;    // an accumulator to normalize the vector elements

 create(HBF::hbf.order());                // allocates the vector elements
 switch (o) {
   case eq_vec :
     factor = one / rp(dimension);           // computes the probability
     for (i=0; i<dimension; i++)             // for all states
       vec[i] = factor;                         // puts the probability
     break;
   case ap_vec :
     for (i=0; i<dimension; i++) {           // for all states
       vec[i] = one / HBF::hbf.coord(i, i);     // puts the inverse of the
       acc += vec[i];                        // corresp. diagonal element
     }                                          // and add to the accumulator
     factor = one / acc;                  // normalizes the vector
     for (i=0; i<dimension; i++)
       vec[i] *= factor;
     break;
 }
}

     //--------------------------------------------------------------------------//
     //                      Output on 'fout' file stream                        //
     //--------------------------------------------------------------------------//
void VCT::write() const
{
 ofstream fout;
 Open_File(file_name, vct_file, fout);
 fout << dimension << "\n";
 sizes.write(fout);
 p_order.write(fout);
 p_jumps.write(fout);
 for (int i=0; i<dimension; i++)
   fout << vec[i] << "\n";
 Close_File(file_name, vct_file, fout);
}
     //--------------------------------------------------------------------------//
     //                      Input from 'fin' file stream                        //
     //--------------------------------------------------------------------------//
void VCT::read(const peps_name file)
{
 ifstream fin;
 int d;

 Open_File(file, vct_file, fin);
 fin >> d;
 create(d);
 baptise(file);
 sizes.read(fin);
 p_order.read(fin);
 p_jumps.read(fin);
 for (d=0; d<dimension; d++)
   fin >> vec[d];
 Close_File(file_name, vct_file, fin);
}
     //--------------------------------------------------------------------------//
     //              Memory space used by 'this' (in memory units)               //
     //--------------------------------------------------------------------------//
int VCT::mem_use() const
{
 int answer = 4;
 answer += dimension * 8;
 answer += p_order.mem_use();
 answer += p_jumps.mem_use();
 answer += sizes.mem_use();
 return(answer);
}


