//====================================================================================//
//                                                                                    //
//                            Class Probability Vector                                //
//                                                                                    //
//====================================================================================//
//  This File:   vct.h                      Language: C++  (xlC and CC)               //
//  Software:    PEPS 3.1                    Advisor: Brigitte.Plateau@imag.fr        //
//  Error Codes: 2300, 2400                                                           //
//====================================================================================//
//  Creation:    19/dec/95                        by: Paulo.Fernandes@imag.fr         //
//  Last Change: 16/feb/00                        by: Paulo.Fernandes@imag.fr         //
//====================================================================================//
//     This class is used to define probability vectors in general. It defines also   //
// the PEPS main structure probability vector (VCT::vct).                             //
//                                                                                    //
//     The variable 'dimension' stores the total number of states of the vector. The  //
// vector elements (stored in the array field 'vec') represent the probability of     //
// each global state. These global states are sorted in a lexical order. The automata //
// list object 'permutation' defines the order of the "letters" in the "word" of this //
// lexical order (see example below). The automata sizes object 'sizes' stores the    //
// size of each automaton.                                                            //
//                                                                                    //
// Note: The permutation list must include all automata of the model.                 //
//                                                                                    //
// Note: The total number of states in the vector must be equal to the product of the //
// sizes of the automata ('dimension' == 'sizes.ss()').                               //
//                                                                                    //
// Example: to this two different 'permutation' lists, the vector elements will       //
// correspond to the following global states ('sizes' = { 2, 3, 4 }):                 //
//                                                                                    //
// +--------------------------------------+  +--------------------------------------+ //
// |     'permutation' = { 0, 1, 2 }      |  |     'permutation' = { 2, 0, 1 }      | //
// +----+-------+----+-------+----+-------+  +----+-------+----+-------+----+-------+ //
// |  0 | 0 0 0 |  8 | 0 2 0 | 16 | 1 1 0 |  |  0 | 0 0 0 |  8 | 0 2 1 | 16 | 1 1 2 | //
// |  1 | 0 0 1 |  9 | 0 2 1 | 17 | 1 1 1 |  |  1 | 0 1 0 |  9 | 1 0 1 | 17 | 1 2 2 | //
// |  2 | 0 0 2 | 10 | 0 2 2 | 18 | 1 1 2 |  |  2 | 0 2 0 | 10 | 1 1 1 | 18 | 0 0 3 | //
// |  3 | 0 0 3 | 11 | 0 2 3 | 19 | 1 1 3 |  |  3 | 1 0 0 | 11 | 1 2 1 | 19 | 0 1 3 | //
// |  4 | 0 1 0 | 12 | 1 0 0 | 20 | 1 2 0 |  |  4 | 1 1 0 | 12 | 0 0 2 | 20 | 0 2 3 | //
// |  5 | 0 1 1 | 13 | 1 0 1 | 21 | 1 2 1 |  |  5 | 1 2 0 | 13 | 0 1 2 | 21 | 1 0 3 | //
// |  6 | 0 1 2 | 14 | 1 0 2 | 22 | 1 2 2 |  |  6 | 0 0 1 | 14 | 0 2 2 | 22 | 1 1 3 | //
// |  7 | 0 1 3 | 15 | 1 0 3 | 23 | 1 2 3 |  |  7 | 0 1 1 | 15 | 1 0 2 | 23 | 1 2 3 | //
// +----+-------+----+-------+----+-------+  +----+-------+----+-------+----+-------+ //
//====================================================================================//

class VCT
{
public: 
 inline VCT();                                                      // Empty Constructor
 inline VCT(const int dim);                                         // Sized Constructor
 inline VCT(const VCT & v);                                          // Copy Constructor
 inline ~VCT();                                                            // Destructor

 inline void create();                                            // Empty Intialization
 inline void create(const int dim);                               // Sized Intialization
 inline void copy(const VCT & v);                        // Copy and Sized Intialization
 inline void erase();                                                     // Destruction

 inline void baptise(const peps_name file);                    // It assigns a file name
 inline void name(peps_name file) const;                     // It returns the file name
 inline int  size() const;                                 // The total number of states
 inline bool is_empty() const;                     // The total number of states is zero

 inline rp   get(const int s) const;                 // The probability of the state 's'
 inline void put(const int s, const rp v);   // It puts probability 'v' to the state 's'


 inline bool test(const VCT & v, const e_typ err, const rp e) const;
               // It returns 'true' if the error between two vectors is smaller than 'e'
 inline void info(rp & acc, rp & big, int & bpos, rp & sml, int & spos) const;
      // It returns the vector sum, the largest and smallest elements and their position

        //--------------------------------------------------------------------------//
        //               It initializes the states of 'this' - HBF                  //
        //--------------------------------------------------------------------------//
        //    This function is used to initialize the elements of 'this' according  //
        // to one of the following options:                                         //
        //                                                                          //
        //   - 'eq_vec',      all reachable states are initialized with an uniform  //
        //                    distribution (equiprobable);                          //
        //   - 'ap_vec',      all reachable states are initialized with a           //
        //                    distribution proportional to the inverse of the       //
        //                    diagonal elements of the HBF matrix ('HBF::hbf');     //
        //                                                                          //
        //   The size of the vector is to the HBF matrix order. The fields 'sizes'  //
        // and 'permutation' are erased.                                            //
        //                                                                          //
        // Note: this functions is used to prepare a vector to be multiplied by the //
        // Hardwell-Boeing Format sparse matrix (static variable 'HBF::hbf').       //
        //--------------------------------------------------------------------------//
        void init_hbf(const v_typ o);

        void write() const;                              // Output on 'fout' file stream
        void read(const peps_name file);                 // Input from 'fin' file stream
        int  mem_use() const;                  // Memory space used by 'this' (in bytes)

        //--------------------------------------------------------------------------//
        //                             Static variable                              //
        //--------------------------------------------------------------------------//
        static VCT vct;
        //--------------------------------------------------------------------------//
        //                             Main Functions                               //
        //--------------------------------------------------------------------------//
        // Basic multiplications                                                    //
        //--------------------------------------------------------------------------//
        void mult_HBF(const VCT & v);    // Multiply by a HBF

        //--------------------------------------------------------------------------//
        //                          Data Structures                                 //
        //--------------------------------------------------------------------------//
private: peps_name file_name;          // the file name of this structure
         int       dimension;          // the number of states (vector size)
         rp      * vec;                // vector elements
         aut_list  p_order;            // the current permutation order of the vector
         jumps     p_jumps;            // the current permutation jumps of the vector
         aut_st    sizes;              // the size of each automaton

};

